/**
 * Semi-automatically generated files used to measure predicate coverage.
 * Details can be found at http://mir.cs.illinois.edu/coverage
 */
package predicate.avl;

public class IntAVLTreeMap_pred {
    private Node root;
    private int size;

    public class Node {
        int key;
        Object value;
        Node left;
        Node right;
        int height;

        public String linearize() {
            String s = "N" + "[" + (left != null ? left.linearize() : null)
                    + "][" + (right != null ? right.linearize() : null) + "]";
            return s;
        }

        @Override
        public String toString() {
            String s = "{c:" + ";v" + value + "["
                    + (left != null ? left.toString() : null) + "]["
                    + (right != null ? right.toString() : null) + "]}";
            return s;
        }

    }

    public String linearize() {
        if (root == null) {
            executedPredicate(0, null);
            return "null";
        } else {
            executedPredicate(1, root);
            return root.linearize();
        }

    }

    public String toString() {
        if (root == null)
            return "null";
        else
            return root.toString();
    }

    public boolean containsKey(Integer key) {
        return findNode(key) != null;
    }

    public Object get(Integer key) {
        Node x = findNode(key);
        return x == null ? null : x.value;
    }

    void put(Integer key, Object value) {
        Node x = new Node();
        x.key = key;
        x.value = value;
        treeInsert(x);
    }

    public Object remove(Integer key) {
        Node fnd = findNode(key);
        if (fnd == null) {
            executedPredicate(2, null);
            return null;
        } else {
            executedPredicate(3, fnd);
        }

        Node ret = treeDelete(fnd);
        size--;
        return ret.value;
    }

    public int getSize() {
        return size;
    }

    private int getHeight(Node x) {
        return x == null ? -1 : x.height;
    }

    private int getBalance(Node x) {
        return x == null ? 0 : getHeight(x.left) - getHeight(x.right);
    }

    private Node findNode(Integer key) {
        Node cur = root;
        while (cur != null && key.compareTo(cur.key) != 0) {
            executedPredicate(4, cur);
            if (key.compareTo(cur.key) < 0) {
                executedPredicate(6, cur);
                cur = cur.left;
            } else {
                executedPredicate(7, cur);
                cur = cur.right;
            }
        }
        executedPredicate(5, cur);
        return cur;
    }

    private Node rightRotate(Node x) {
        if (x == null) {
            executedPredicate(8, null);
            return null;
        } else {
            executedPredicate(9, x);
        }

        Node y = x.left;
        x.left = y.right;
        y.right = x;
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        y.height = 1 + Math.max(getHeight(y.left), getHeight(x));
        return y;
    }

    private Node leftRotate(Node x) {
        if (x == null) {
            executedPredicate(10, null);
            return null;
        } else {
            executedPredicate(11, x);
        }

        Node y = x.right;
        x.right = y.left;
        y.left = x;
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        y.height = 1 + Math.max(getHeight(y.right), getHeight(x));
        return y;
    }

    private Node rightLeftRotate(Node x) {
        x.right = rightRotate(x.right);
        return leftRotate(x);
    }

    private Node leftRightRotate(Node x) {
        x.left = leftRotate(x.left);
        return rightRotate(x);
    }

    public void treeInsert(Node x) {
        Node fnd = findNode(x.key);
        if (fnd != null) {
            executedPredicate(12, fnd);
            fnd.key = x.key;
            return;
        } else {
            executedPredicate(13, null);
        }

        root = treeInsert_recur(x, root);
        size++;
    }

    private Node treeInsert_recur(Node x, Node cur) {
        if (cur == null) {
            executedPredicate(14, cur);
            return x;
        } else {
            executedPredicate(15, null);
        }

        if (((Integer) x.key).compareTo(cur.key) < 0) {
            executedPredicate(16, cur);
            cur.left = treeInsert_recur(x, cur.left);
        } else {
            executedPredicate(17, cur);
            if (((Integer) x.key).compareTo(cur.key) == 0) {
                executedPredicate(18, cur);
                cur.value = x.value;
                return cur;
            } else {
                executedPredicate(19, cur);
                cur.right = treeInsert_recur(x, cur.right);
            }
        }

        cur.height = 1 + Math.max(getHeight(cur.left), getHeight(cur.right));
        int balance = getBalance(cur);
        if (balance == -2) {
            executedPredicate(20, cur);
            {
                if (getBalance(cur.right) == -1) {
                    executedPredicate(22, cur);
                    return leftRotate(cur);
                } else {
                    executedPredicate(23, cur);
                    if (getBalance(cur.right) == 1) {
                        executedPredicate(24, cur);
                        return rightLeftRotate(cur);
                    } else {
                        executedPredicate(25, cur);
                    }
                }

            }
        } else {
            executedPredicate(21, cur);
            if (balance == 2) {
                executedPredicate(26, cur);
                if (getBalance(cur.left) == 1) {
                    executedPredicate(28, cur);
                    return rightRotate(cur);
                } else {
                    executedPredicate(29, cur);
                    if (getBalance(cur.left) == -1) {
                        executedPredicate(30, cur);
                        return leftRightRotate(cur);
                    } else {
                        executedPredicate(31, cur);
                    }
                }
            } else {
                executedPredicate(27, cur);
            }
        }

        return cur;
    }

    private Node treeDelete(Node x) {
        root = treeDelete_recur(x, root);
        return x;
    }

    private Node treeDelete_recur(Node x, Node cur) {
        if (((Integer) x.key).compareTo(cur.key) < 0) {
            executedPredicate(32, cur);
            cur.left = treeDelete_recur(x, cur.left);
        } else {
            executedPredicate(33, cur);
            if (((Integer) x.key).compareTo(cur.key) == 0) {
                executedPredicate(34, cur);
                return afterDelete(cur);
            } else {
                executedPredicate(35, cur);
                cur.right = treeDelete_recur(x, cur.right);
            }
        }

        return DeleteFix(cur);
    }

    private Node DeleteFix(Node cur) {
        cur.height = 1 + Math.max(getHeight(cur.left), getHeight(cur.right));
        int balance = getBalance(cur);
        if (balance == -2) {
            executedPredicate(36, cur);
            if (getBalance(cur.right) <= 0) {
                executedPredicate(38, cur);
                return leftRotate(cur);
            } else {
                executedPredicate(39, cur);
                if (getBalance(cur.right) == 1) {
                    executedPredicate(40, cur);
                    return rightLeftRotate(cur);
                } else {
                    executedPredicate(41, cur);
                }
            }
        } else {
            executedPredicate(37, cur);
            if (balance == 2) {
                executedPredicate(42, cur);
                if (getBalance(cur.left) >= 0) {
                    executedPredicate(44, cur);
                    return rightRotate(cur);
                } else {
                    executedPredicate(45, cur);
                    if (getBalance(cur.left) == -1) {
                        executedPredicate(46, cur);
                        return leftRightRotate(cur);
                    } else {
                        executedPredicate(47, cur);
                    }
                }
            } else {
                executedPredicate(43, cur);
            }
        }

        return cur;
    }

    private Node afterDelete(Node x) {
        if (x.left == null && x.right == null) {
            executedPredicate(48, x);
            return null;
        } else {
            executedPredicate(49, x);
            if (x.left != null && x.right != null) {
                executedPredicate(50, x);
                Node z = getIOS(x);
                x.right = treeDelete_recur(z, x.right);
                x.key = z.key;
                return DeleteFix(x);
            } else {
                executedPredicate(51, x);
                if (x.left != null) {
                    executedPredicate(52, x);
                    return x.left;
                } else {
                    executedPredicate(53, x);
                    return x.right;
                }
            }
        }

    }

    private Node getIOS(Node z) {
        Node x, y = null;
        x = z.right;
        while (x != null) {
            executedPredicate(54, x);
            y = x;
            x = x.left;
        }
        executedPredicate(55, z);
        return y;
    }

    protected int compare(Integer k1, Integer k2) {
        return k1.compareTo(k2);
    }

    private static int NUM_OF_PREDICATES = 4;

    public static int numOfPredicates() {
        return NUM_OF_PREDICATES;
    }

    public void executedPredicate(int branch_id, Node n) {
        int res = 0;
        if (n != null) {
            res |= 1 << 0;
            res |= (n.left == null) ? 0 : 1 << 1;
            res |= (n.right == null) ? 0 : 1 << 2;
            res |= (n == root) ? 0 : 1 << 3;
        }
        System.out.println("branchId=" + branch_id + "; covered predicate combination=" + res);
    }

}
