/**
 * Semi-automatically generated files used to measure predicate coverage.
 * Details can be found at http://mir.cs.illinois.edu/coverage
 */
package predicate.fibonacciheap;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

public class FibonacciHeap_pred {

    private final static double oneOverLogPhi = 1.0 / Math.log((1.0 + Math
            .sqrt(5.0)) / 2.0);

    private final static int FALSE = 0;
    private final static int TRUE = 1;

    public FibonacciHeapNode minNode;

    public int size;

    public void delete(FibonacciHeapNode x) {
        // make x as small as possible
        decreaseKey(x, Integer.MIN_VALUE);

        // remove the smallest, which decreases n also
        removeMin();
    }

    public void decreaseKey(FibonacciHeapNode x, int k) {
        if (k > x.key) {
            executedPredicate(0, x, null);
            throw new IllegalArgumentException(
                    "decreaseKey() got larger key value");
        } else {
            executedPredicate(1, x, null);
        }

        x.key = k;
        FibonacciHeapNode y = x.parent;
        if ((y != null) && (x.key < y.key)) {
            executedPredicate(2, x, y);
            cut(x, y);
            cascadingCut(y);
        } else {
            executedPredicate(3, x, null);
        }

        if (x.key < minNode.key) {
            executedPredicate(4, x, minNode);
            minNode = x;
        } else {
            executedPredicate(5, x, minNode);
        }

    }

    protected void cascadingCut(FibonacciHeapNode y) {
        FibonacciHeapNode z = y.parent;
        if (z != null) {
            executedPredicate(6, z, null);
            if (y.mark == FALSE) {
                executedPredicate(8, y, null);
                y.mark = TRUE;
            } else {
                executedPredicate(9, y, null);
                cut(y, z);
                cascadingCut(z);
            }
        } else {
            executedPredicate(7, y, null);
        }

    }

    protected void cut(FibonacciHeapNode x, FibonacciHeapNode y) {
        x.left.right = x.right;
        x.right.left = x.left;
        y.degree--;
        if (y.child == x) {
            executedPredicate(10, x, y);
            y.child = x.right;
        } else {
            executedPredicate(11, x, y);
        }

        if (y.degree == 0) {
            executedPredicate(12, y, x);
            y.child = null;
        } else {
            executedPredicate(13, x, y);
        }

        x.left = minNode;
        x.right = minNode.right;
        minNode.right = x;
        x.right.left = x;
        x.parent = null;
        x.mark = FALSE;
    }

    public FibonacciHeapNode removeMin() {
        FibonacciHeapNode z = minNode;
        if (z != null) {
            executedPredicate(14, z, null);
            int numKids = z.degree;
            FibonacciHeapNode x = z.child;
            FibonacciHeapNode tempRight;
            while (numKids > 0) {
                executedPredicate(16, x, z);
                tempRight = x.right;
                x.left.right = x.right;
                x.right.left = x.left;
                x.left = minNode;
                x.right = minNode.right;
                minNode.right = x;
                x.right.left = x;
                x.parent = null;
                x = tempRight;
                numKids--;
            }
            executedPredicate(17, x, z);
            z.left.right = z.right;
            z.right.left = z.left;
            if (z == z.right) {
                executedPredicate(18, x, z);
                minNode = null;
            } else {
                executedPredicate(19, x, z);
                minNode = z.right;
                consolidate();
            }

            size--;
        } else {
            executedPredicate(15, z, null);
        }

        return z;
    }

    protected void consolidate() {
        int arraySize = ((int) Math.floor(Math.log(size) * oneOverLogPhi)) + 1;
        List<FibonacciHeapNode> array = new ArrayList<FibonacciHeapNode>(
                arraySize);
        for (int i = 0; i < arraySize; i++) {
            executedPredicate(20, null, null);
            array.add(null);
        }
        executedPredicate(21, null, null);

        int numRoots = 0;
        FibonacciHeapNode x = minNode;
        if (x != null) {
            executedPredicate(22, null, null);
            numRoots++;
            x = x.right;
            while (x != minNode) {
                executedPredicate(24, x, null);
                numRoots++;
                x = x.right;
            }
            executedPredicate(25, x, null);
        } else {
            executedPredicate(23, minNode, null);
        }

        while (numRoots > 0) {
            executedPredicate(26, minNode, null);
            int d = x.degree;
            FibonacciHeapNode next = x.right;
            for (;;) {
                executedPredicate(28, array.get(d), null);
                FibonacciHeapNode y = array.get(d);
                if (y == null) {
                    executedPredicate(30, x, y);
                    break;
                } else {
                    executedPredicate(31, x, y);
                }

                if (x.key > y.key) {
                    executedPredicate(32, x, y);
                    FibonacciHeapNode temp = y;
                    y = x;
                    x = temp;
                } else {
                    executedPredicate(33, x, null);
                }

                link(y, x);
                array.set(d, null);
                d++;
            }
            executedPredicate(29, minNode, null);

            array.set(d, x);
            x = next;
            numRoots--;
        }
        executedPredicate(27, minNode, null);
        minNode = null;
        for (int i = 0; i < arraySize; i++) {
            executedPredicate(34, array.get(i), null);
            FibonacciHeapNode y = array.get(i);
            if (y == null) {
                executedPredicate(36, array.get(i), null);
                continue;
            } else {
                executedPredicate(37, array.get(i), null);
            }

            if (minNode != null) {
                executedPredicate(38, array.get(i), null);
                y.left.right = y.right;
                y.right.left = y.left;
                y.left = minNode;
                y.right = minNode.right;
                minNode.right = y;
                y.right.left = y;
                if (y.key < minNode.key) {
                    executedPredicate(40, array.get(i), minNode);
                    {
                        minNode = y;
                    }
                } else {
                    executedPredicate(41, array.get(i), minNode);
                }
            } else {
                executedPredicate(39, array.get(i), null);
                minNode = y;
            }
        }
        executedPredicate(35, minNode, null);
    }

    protected void link(FibonacciHeapNode y, FibonacciHeapNode x) {
        y.left.right = y.right;
        y.right.left = y.left;
        y.parent = x;
        if (x.child == null) {
            executedPredicate(42, y, x);
            x.child = y;
            y.right = y;
            y.left = y;
        } else {
            executedPredicate(43, y, x);
            y.left = x.child;
            y.right = x.child.right;
            x.child.right = y;
            y.right.left = y;
        }

        x.degree++;
        y.mark = FALSE;
    }

    public void insert(FibonacciHeapNode node, int key) {
        node.key = key;
        if (minNode != null) {
            executedPredicate(44, minNode, null);
            node.left = minNode;
            node.right = minNode.right;
            minNode.right = node;
            node.right.left = node;
            if (key < minNode.key) {
                executedPredicate(46, minNode, null);
                minNode = node;
            } else {
                executedPredicate(47, minNode, null);
            }
        } else {
            executedPredicate(45, minNode, null);
            minNode = node;
        }

        size++;
    }

    private int getMin() {
        FibonacciHeapNode temp = minNode;
        int min = minNode.key;
        do {
            if (temp.key < min) {
                executedPredicate(48, minNode, temp);
                min = temp.key;
            } else {
                executedPredicate(49, minNode, temp);
            }

            temp = temp.right;
        } while (temp != minNode);
        return min;
    }

    public String toString() {
        if (minNode == null) {
            executedPredicate(50, minNode, null);
            return "FibonacciHeap=[]";
        } else {
            executedPredicate(51, minNode, null);
        }

        Stack<FibonacciHeapNode> stack = new Stack<FibonacciHeapNode>();
        stack.push(minNode);
        StringBuffer buf = new StringBuffer(512);
        buf.append("FibonacciHeap=[");
        while (!stack.empty()) {
            executedPredicate(52, minNode, null);
            FibonacciHeapNode curr = stack.pop();
            buf.append(curr + (curr == minNode ? "MIN" : ""));
            buf.append(", ");
            if (curr.child != null) {
                executedPredicate(54, minNode, curr);
                stack.push(curr.child);
            } else {
                executedPredicate(55, minNode, curr);
            }

            FibonacciHeapNode start = curr;
            curr = curr.right;
            while (curr != start) {
                executedPredicate(56, curr, start);
                buf.append(curr + (curr == minNode ? "MIN" : ""));
                buf.append(", ");
                if (curr.child != null) {
                    executedPredicate(58, curr, start);
                    stack.push(curr.child);
                } else {
                    executedPredicate(59, curr, start);
                }

                curr = curr.right;
            }
            executedPredicate(57, minNode, null);
        }
        executedPredicate(53, minNode, null);
        buf.append(']');
        return buf.toString();
    }

    public boolean checkHeapified() {
        FibonacciHeapNode current = minNode;
        do {
            if (!current.checkHeapified()) {
                executedPredicate(60, current, null);
                return false;
            } else {
                executedPredicate(61, current, null);
            }

            current = current.right;
        } while (current != minNode);
        return true;
    }

    private static int NUM_OF_PREDICATES = 14;

    public static int numOfPredicates() {
        return NUM_OF_PREDICATES;
    }

    public void executedPredicate(int branch_id, FibonacciHeapNode n,
            FibonacciHeapNode m) {
        int res = 0;

        if (n != null) {
            res |= 1 << 0;
            res |= (n.child == null) ? 0 : 1 << 1;
            res |= (n.parent == null) ? 0 : 1 << 2;
            res |= (n.right == null) ? 0 : 1 << 3;
            res |= (n.left == null) ? 0 : 1 << 4;
            res |= (n.degree == 0) ? 0 : 1 << 5;
        }
        if (m != null) {
            res |= 1 << 6;
            res |= (m.child == null) ? 0 : 1 << 7;
            res |= (m.parent == null) ? 0 : 1 << 8;
            res |= (m.right == null) ? 0 : 1 << 9;
            res |= (m.left == null) ? 0 : 1 << 10;
            res |= (m.degree == 0) ? 0 : 1 << 11;
        }
        if (n != null && m != null) {
            res |= (n.child == m) ? 0 : 1 << 12;
            res |= (m.child == n) ? 0 : 1 << 13;
        }
        System.out.println("branchId=" + branch_id + "; covered predicate combination=" + res);
    }

}
